#include <stdio.h>
#include <fcntl.h>
#include <errno.h>
#include <unistd.h>

#include "pe.h"

/* PREFERRED BASE ADDRESS FOR THE IMAGE */
#define PREFER_ADD    0x400000
/* Total number of arguments to this utility.*/
#define ARGC    3

/* Executable File argument index in argv*/

#define EXECUTABLE 		1
#define IMAGE 			2
#define dprintf   		printf
#define NUM_MAX_SECTION 	96
#define	BUFLEN			4096


int		make_exec_info();	
int 		extract_image(int fd, int fd1, int i);

coff_header		cheader;
sectn_header		sheaders[NUM_MAX_SECTION];
optnl_header    	opt_hdr;
int 			fd, fd1;
unsigned long 		relative_base;
unsigned long		data_start = 0;
unsigned long		data_end = 0;

usage( char *name)
{
	printf("Usage %s <executable name>  <image file name>\n", name);
}

main(int argc, char **argv)
{
	int err = 0;

	if (argc != ARGC) {
		usage(argv[0]);
		exit (1);
	}
	if ((fd = open(argv[EXECUTABLE], O_RDONLY)) < 0) {
		printf("Error opening the executable %s.  Exiting.\n",argv[EXECUTABLE]);
		exit(2);
	}
	if ((fd1 = open(argv[IMAGE], O_RDWR|O_CREAT|O_TRUNC)) < 0) {
		printf("Error creating image file %s.  Exiting\n",argv[IMAGE]);
		exit(2);
	}
#ifdef DEBUG
	dprintf("Starting creating image file.\n");
#endif
	if (err = make_exec_info()) {
		printf("Error: %x\n", err);
		exit(3);
	}
#ifndef __BEOS__
	if (fchmod(fd1,0777) < 0) {
		printf("Error setting r/w/x permissions on image file.\n");
		exit (4);
	}
#endif
	exit(0);
}

int
make_exec_info()
{
	int			err;
	int			size;
	int			i;
	char			pe_buf[SIZE_OF_PE_SIGNATURE];
	int			j;
	

	/* At this point it is assumed that both fd and fd1 are valid
	 * file descriptors.  And no read/write operation is yet done on it.
	 */
	err = read(fd, pe_buf, 2);
	if (err != 2) {
		dprintf("Not able to read executable\n");
		err = ENOEXEC;
		goto error1;
	}
	if ((pe_buf[0] != 'M') || (pe_buf[1] != 'Z')) {
		dprintf("DOS header signature mismatch.\n");
		err = ENOEXEC;
		goto error1;
	}

	lseek(fd, PE_SIGNATURE_OFFSET, 0);
	err = read(fd, pe_buf, 4);
	if (err != 4) {
#ifdef DEBUG
		dprintf("PE: couldn't read PE signature offset.\n");
#endif
		err = ENOEXEC;
		goto error1;
	}
	i = *((int *)pe_buf);

	lseek(fd, i, 0);

	err = read(fd, pe_buf, SIZE_OF_PE_SIGNATURE);
	if (err != SIZE_OF_PE_SIGNATURE) {
#ifdef DEBUG
		dprintf("PE: couldn't read PE signature.\n");
#endif
		err = ENOEXEC;
		goto error1;
	}

	i = 0;
	if (pe_buf[i] == 'P') {
		if (pe_buf[i+1] != 'E') {
#ifdef DEBUG
			dprintf("PE: PE signature-E mismatch.\n");
#endif
			err = ENOEXEC;
			goto error1;
		}
	} else  {
#ifdef DEBUG
			dprintf("PE: PE signature-P mismatch.\n");
#endif
		err = ENOEXEC;
		goto error1;
	}

	err = read(fd, &cheader, sizeof(coff_header));
	if (err != sizeof(coff_header)) {
#ifdef DEBUG
			dprintf("PE: couldn't read coff header.\n");
#endif
		err = ENOEXEC;
		goto error1;
	}

	if (cheader.mchn_typ != 0x14c) { /*Not for i386 */
#ifdef DEBUG
		dprintf("PE: machine type mismatch-read %x\n", cheader.mchn_typ);
#endif
		err = ENOEXEC;
		goto error1;
	}
	if (cheader.num_sectns > NUM_MAX_SECTION) {
#ifdef DEBUG
		dprintf("PE: No. of sections exceeds limit %x\n", cheader.num_sectns);
#endif
		err = ENOEXEC;
		goto error1;
	}
	if (cheader.opthdr_sz == 0) { /*Doesn't have optional header*/
#ifdef DEBUG
		dprintf("PE: Does not have optional header\n");
#endif
		err = ENOEXEC;
		goto error1;
	}

	/*
	 * Reading optional header.
	 */
	err = read(fd, &opt_hdr, sizeof(optnl_header));
	if (err != sizeof(optnl_header)) {
#ifdef DEBUG
		dprintf("PE: Error reading optional header\n");
#endif
		err = ENOEXEC;
		goto error1;
	}
	if (opt_hdr.std.magic != EXEC_FILE)  {
#ifdef DEBUG
		dprintf("PE: Magic number mismatch %x\n", opt_hdr.std.magic);
#endif
		err = ENOEXEC;
		goto error1;
	}
	/*
	 *	Reading Section headers.
	 */
	err = read(fd, &sheaders[0], cheader.num_sectns * 
	                                  sizeof(sectn_header));
	if (err != cheader.num_sectns * sizeof(sectn_header)) {
#ifdef DEBUG
		dprintf("PE: Error reading Sections headers\n");
#endif
		err = ENOEXEC;
		goto error1;
	}

	/* All the sections in image file will be stored w.r.t. text rva.
	 */
	relative_base = (unsigned long)sheaders[0].rva;
	for (i=0;i<cheader.num_sectns;i++)
		if (err = extract_image(fd, fd1, i)) {
			dprintf("Error reading section %d\n", i);
			goto error1;
		}
	return 0;

error1:
	return err;
}

int
extract_image(int fd, int fd1, int i)
{
	sectn_header	 *sheader;
	long		size;
	int 		err;
	char		buf[BUFLEN];
	int			temp, tempsize;

	/* read in segment. 
	 */
	sheader = sheaders + i;
	if (sheader->virtualsz > sheader->raw_datasz)
		 /* For sections containing all zeros, since 
		  * we should read only this much from disk 
		  * and PAD the rest with zeros later.
		  */
		size = sheader->raw_datasz;
	else
		/* For File alignment sections */
		size = sheader->virtualsz;

	lseek(fd, sheader->raw_datap, SEEK_SET);
	lseek(fd1, (off_t)(sheader->rva - relative_base), SEEK_SET);

	if (size != 0) {
		temp = size;
		while (temp) {
			if (temp > BUFLEN)  {
				tempsize = BUFLEN;
				temp -= BUFLEN;
			}
			else {
				tempsize = temp;
				temp = 0;
			}
			err = read(fd, buf, tempsize);
			if (err != tempsize) {
				dprintf("Error reading executable file. Exiting!\n");
				err =  ENOEXEC;
				goto error1;
			}
			err = write(fd1, buf, tempsize);
			if (err != tempsize) {
				dprintf("Error writing image file. Exiting!\n");
				err =  ENOEXEC;
				goto error1;
			}
		}
	}
#ifdef DEBUG
	dprintf("Section %d, Bytes read %x\n", i, err);
#endif

	/* zero the end of the segment only if needed*/
	memset(buf, 0, BUFLEN);


#if 0
	if (size == 0) {
		/*
		 * .bss section fill in the data_start and data_end variables
		 */

		*(ulong *)&buf[0] = (ulong)opt_hdr.std.data_base;
		*(ulong *)&buf[4] = (ulong)opt_hdr.std.data_base + (opt_hdr.std.dsize + opt_hdr.std.bsize);
	}
#endif
	temp = sheader->virtualsz - size;
	while (temp > 0) {
		if (temp > BUFLEN) {
			tempsize = BUFLEN;
			temp -= BUFLEN;
		}
		else {
			tempsize = temp;
			temp = 0;
		}
		err = write(fd1, buf, tempsize);
		if (err != tempsize) {
			dprintf("Error writing zeros in image file. Exiting!\n");
			err =  ENOEXEC;
			goto error1;
		}
	}


	/*
	 * fixup the data_start and data_end
	 */

	if (!strcmp(sheader->name, ".data")) { 
		*(ulong *)&buf[0] = (ulong)opt_hdr.std.data_base + PREFER_ADD - 0x1000;
		*(ulong *)&buf[4] = PREFER_ADD - 0x1000 + 
		                    (ulong)opt_hdr.std.data_base + 
							(opt_hdr.std.dsize + opt_hdr.std.bsize);

		dprintf("BASE %x SIZE %x\n", *(ulong *)&buf[0], *(ulong *)&buf[4]);
		lseek(fd1, (off_t)(sheader->rva - relative_base), SEEK_SET);
		err = write(fd1, buf, 8);
		if (err != 8) {
			dprintf("Error in fixing up data start and data end\n");
			err =  ENOEXEC;
			goto error1;
		}
	}
		
	return 0;

error1:
	printf("Error reading the section.\n");
	return err;
}
